#!/usr/bin/env python
# coding: utf-8

# In[7]:


# 日本語フォント環境のインストール（import japaneize_matplotlibでエラーが出た場合に，次の行の#を消して実行）
#!pip install japanize-matplotlib

import numpy as np
import matplotlib.pyplot as plt
import japanize_matplotlib

class Quad_Plot_Fn:
    """
    2次関数を描画するクラス
    """
    def __init__(self):
        """ 凡例の係数フォーマット関数 """
        self.eqlabel = Graph_Eqlabel() 

        self._ptsize = 100  # 点の大きさ
        self._axis_linewidth = 2 # 軸の線幅
        
    def quad_plot_vertex(self, a, p, q, x_range, linestyle, color):
        """ 2次関数のグラフ描画（頂点形式）"""
        # 放物線の描画値を計算 (x_min ≦ x ≦ x_max)
        x_min, x_max = x_range
        x = np.linspace(x_min, x_max, 100)
        y = a * (x - p)**2 + q  # 2次関数 y = a(x - p)^2 + q の計算
        
        # 2次関数のグラフを描画
        plt.plot(x, y, linestyle=linestyle, color=color, label = self.eqlabel.quad_vrtx(a, p, q))

    def quad_plot_standard(self, a, b, c, x_range, linestyle, color):
        """ 2次関数のグラフ描画（標準形式）"""
        # 放物線の描画値を計算 (x_min ≦ x ≦ x_max)
        x_min, x_max = x_range
        x = np.linspace(x_min, x_max, 100)
        y = a * x**2 + b * x + c  # 2次関数 y = ax^2 + bx + c の計算
        
        # 2次関数のグラフを描画
        plt.plot(x, y, linestyle=linestyle, color=color, label = self.eqlabel.quad_std(a, b, c))

    def peak_axis_plot(self, p):
        """ 対称軸（頂点を通る縦の線）を描画 """
        plt.axvline(x=p, color='red', linestyle='--', linewidth = self._axis_linewidth, label=f'軸: x = {p}')

    def pt_plot(self, pt):
        """ 座標点のプロット """
        plt.scatter(pt[0], pt[1], color='red', s = self._ptsize, zorder=5)  

class Linear_Plot_Fn:
    """
    1次関数を描画するクラス
    """
    def __init__(self):
        """ 凡例の係数フォーマット関数 """
        self.eqlabel = Graph_Eqlabel() 

        self._ptsize = 100  # 点の大きさ
        self._fontsize = 18 # ラベルサイズ
        
    def linear_plot(self, a, b, x_range, linestyle):
        # 直線の描画値を計算 (x_min ≦ x ≦ x_max)
        x_min, x_max = x_range
        x = np.linspace(x_min, x_max, 100)
        y = a * x + b  # 1次関数 y = ax + b の計算
    
        # 1次関数のグラフを描画
        plt.plot(x, y, linestyle=linestyle, color='blue', label = self.eqlabel.lin(a, b))

    def x_def_range_drawing(self, a, b, x_def_range):
        x_def_min, x_def_max = x_def_range
        """ 定義域 x_def_min≦x≦x_def_max の範囲で赤い実線のグラフを描画 """
        # y = ax + b の値を赤線で描画
        x_red = np.linspace(x_def_min, x_def_max, 100)
        y_red = a * x_red + b
        plt.plot(x_red, y_red, linestyle='-', color='red', label= self.eqlabel.range_def(x_def_range))
        # 軸上の定義域・値域を赤線で描画
        x_axis_red = np.linspace(0, 0, 100)
        y_axis_red = np.linspace(0, 0, 100)
        plt.plot(x_red, y_axis_red, linestyle='-', color='red', linewidth = 3)
        plt.plot(x_axis_red, y_red, linestyle='-', color='red', linewidth = 3)

    def pt_plot(self, pt):
        """ 座標点のプロット """
        plt.scatter(pt[0], pt[1], color='red', s = self._ptsize, zorder=5)  

    def pt_labelling(self, a, pt, label):
        """ 点のラベルを印字 """
        if a > 0:
            ha_pos = 'left'
        else:
            ha_pos = 'right'
        plt.text(pt[0], pt[1] - 0.1, label , va='top', ha=ha_pos, fontsize=self._fontsize)  

class QuadDefRange:
    """
    2次関数の定義域と対応する関数値を描画し，ラベルを印字するクラス
    """
    def __init__(self):
        """ 係数フォーマット関数 """
        self.eqlabel = Graph_Eqlabel() 
        self.pl = Point_Labelling()
      
        self._ptsize = 100  # 点の大きさ
        self._fontsize = 18 # ラベルサイズ
        self._x_def_linewidth = 2 # 定義域の線幅

    def yvalue_in_def_range(self, a, b, c, x_def_range):
        """ 定義域 x_def_min ≦ x ≦ x_def_max に対する関数の値を赤い実線で描画 """
        x_def_min, x_def_max = x_def_range
        x_red = np.linspace(x_def_min, x_def_max, 100)
        y_red = a*x_red**2 + b*x_red + c
        plt.plot(x_red, y_red, linestyle='-', color='red', label=self.eqlabel.range_def(x_def_range))

    def xvalue_in_def_range(self, x_def_range):
        """ 定義域に対応する x 軸上の範囲を赤線で表示 """
        x_def_min, x_def_max = x_def_range
        plt.plot([x_def_min, x_def_max], [0, 0], color='red', linewidth=self._x_def_linewidth)
    
    def max_min_point_plot_and_labelling(self, a, b, c, x_def_range):
        """ 最大値・最小値の点をプロットして，ラベルを印字 """
        x_def_min, x_def_max = x_def_range
        """ 頂点座標を計算 """
        p = -b / (2 * a)
        q = -b**2/(4 * a) + c
    
        """ 最大・最小のyを持つデータペアを見つける """
        if (x_def_min < p) and (x_def_max > p): # 頂点が定義域内にある場合
            data_pairs = [(x_def_min, a*x_def_min**2 + b*x_def_min + c), (x_def_max, a*x_def_max**2 + b*x_def_max + c), (p, q)]    
        else: # 頂点が定義域内に無い場合
            data_pairs = [(x_def_min, a*x_def_min**2 + b*x_def_min + c), (x_def_max, a*x_def_max**2 + b*x_def_max + c)]
        max_pair = max(data_pairs, key=lambda pair: pair[1])
        min_pair = min(data_pairs, key=lambda pair: pair[1])
        
        """ 最小値 and 最大値を赤い点で表示 """
        plt.scatter(min_pair[0], min_pair[1], color='red', s = self._ptsize, zorder=5)
        plt.scatter(max_pair[0], max_pair[1], color='red', s = self._ptsize, zorder=5)
        
        """ 「最小値」「最大値」のラベル付け（位置を微調整して表示）"""
        self.pl.pt(a, p, min_pair, '最小値')
        self.pl.pt(a, p, max_pair, '最大値')

class QuadEqIdentifying:
    """
    3点のデータから2次関数を決定するクラス
    """
    def __init__(self):
        """ 係数フォーマット関数 """
        self.pl = Point_Labelling()
      
        self._ptsize = 100  # 点の大きさ
        
    def eq_identifying_from_3pts(self, pt1, pt2, pt3):
        x1, y1 = pt1
        x2, y2 = pt2
        x3, y3 = pt3
        
        """ 3点の座標計算 """
        # 方程式を解くための行列を作成
        A = np.array([
            [x1**2, x1, 1],
            [x2**2, x2, 1],
            [x3**2, x3, 1]
        ])
        B = np.array([y1, y2, y3])
        # 連立方程式を解いて a, b, c を求める
        a, b, c = np.linalg.solve(A, B)
    
        # 頂点座標を計算（小数点以下３桁目を四捨五入）
        p = round(-b / (2 * a),2)
        q = round(a * p**2 + b * p + c,2)
        a = round(a,2)

        return a, p, q

    def pt_plotting_and_labelling(self, a, p, pt):
        x, y = pt
        """ 点を赤い点で表示 """
        plt.scatter(x, y, color='red', s = self._ptsize, zorder=5)
    
        """ 点のラベル付け（位置を微調整して表示）"""
        self.pl.pt(a, p, pt, rf'点({x},{y})')


class QuadEqSol:
    """ 
    2次方程式の解を求めるクラス 
    """
    def __init__(self):
        """ 係数フォーマット関数 """
        self.pl = Point_Labelling()
      
        self._ptsize = 100  # 点の大きさ
    
    def quad_eq_sol_fn(self, a, b, c):
        """ 2次方程式の解を計算 """
        roots_x = np.roots([a, b, c])  # 2次方程式の解を求める
        # 複素数解／重解の判定と，重解の場合の丸め誤差処理（重解の場合に丸め誤差により1e-9のオーダーの虚部が残ることがある．）
        if np.abs(roots_x[0].imag) > 1e-5:
            plt.text(0, 5, "解は複素数を含んでいます。", color = 'red', fontsize = 12)
            cross_pts = 'N' # x軸との共有点なし
        elif np.isclose(roots_x[0], roots_x[1]): # 重解を判定
            roots_x = roots_x.real  # 実数解のみ取得
            cross_pts = 'S' # 重解なのでx軸との共有点１つ
        else:
            cross_pts = 'D' # 共有点２つ
        return roots_x, cross_pts

    def draw_crossing_pts(self, a, b, roots_x, cross_pt):
        """ x軸との共有点を描画 （'D':共有点2つ，'S':共有点1つ）"""
        p = -b/2/a # 頂点のx座標
        if cross_pt in ['S', 'D']:
            plt.scatter(roots_x, (0, 0), edgecolor='red', facecolor='red', s=self._ptsize, zorder=5)  # 赤丸をプロット
            self.pl.sol(a, p, (roots_x[0], 0),  rf'({round(roots_x[0],2)},0)')
            if cross_pt == 'D':
                self.pl.sol(a, p, (roots_x[1], 0),  rf'({round(roots_x[1],2)},0)')

class Form_Num:
    """ 
    凡例の係数フォーマットを行うクラス 
    """
    def form_num1(self, value):
        """ 先頭項表記のための係数フォーマット """
        if value == 1:
            return f" "         # 1は表示しない (1x を x)
        elif value == -1:
            return f"-"         # -1は符号だけ (-1x を - x)
        else:
            return f"{value}"   # その他はそのまま (2x は 2x のまま，-2x は -2xのまま)

    def form_num2(self, value):
        """ 第2項表記のための係数フォーマット """
        if value == 1:
            return f"+"         # 1は表示しない (1x を + x)
        elif value == -1:
            return f"-"         # -1は符号だけ (-1x を - x)
        elif value >= 0:
            return f"+{value}"  # 2x は　+2x
        else:
            return f"{value}"   # その他はそのまま (-2x は -2xのまま)
        
    def form_num3(self, value):
        """ 定数項表記のための係数フォーマット """
        if value >= 0:
            return f"+{value}"  # 正の数には + を付ける　(1 を + 1)
        else:
            return f"{value}"   # 負の数はそのまま (- 1 は - 1のまま）


class Graph_Eqlabel:
    """ 
    凡例の式を作成するクラス
    """
    def __init__(self):
        """ 係数フォーマット関数 """
        self.fn = Form_Num() 

    def lin(self, a, b):
        """ 凡例のための1次式を作成 """
        if b == 0:
            return rf'$y = {self.fn.form_num1(a)}x$'                  # y = axと表記する．こうしないと y = ax + 0 と表示されてしまうため．
        else:
            return rf'$y = {self.fn.form_num1(a)}x {self.fn.form_num3(b)}$'   # y = ax + b と表記する．

    def quad_vrtx(self, a, p, q):
        """ 凡例のための2次式（頂点形）を作成 """
        if p == 0 and q == 0:
            return rf'$y = {self.fn.form_num1(a)}x^2$'                # y = ax^2と表記する．こうしないと y = a(x - 0)^2 + 0 と表示されてしまうため．
        elif p == 0:
            return rf'$y = {self.fn.form_num1(a)}x^2 {self.fn.form_num3(q)}$' # y = ax^2 + qと表記する．
        elif q == 0:
            return rf'$y = {self.fn.form_num1(a)}(x {self.fn.form_num3(- p)})^2$' # y = a(x - p)^2 と表記する．
        else:
            return rf'$y = {self.fn.form_num1(a)}(x {self.fn.form_num3(- p)})^2 {self.fn.form_num3(q)}$' # y = a(x - p)^2 + q と表記する．

    def quad_std(self, a, b, c):
        """ 凡例のための2次式（標準形）を作成 """
        if b == 0 and c == 0:
            return rf'$y = {self.fn.form_num1(a)}x^2$'                # y = ax^2と表記する．こうしないと y = ax^2 + 0x + 0 と表示されてしまうため．
        elif b == 0:
            return rf'$y = {self.fn.form_num1(a)}x^2 {self.fn.form_num3(c)}$' # y = ax^2 + cと表記する．
        elif c == 0:
            return rf'$y = {self.fn.form_num1(a)}x^2 {self.fn.form_num2(b)}x$' # y = ax^2 + bx  と表記する．
        else:
            return rf'$y = {self.fn.form_num1(a)}x^2 {self.fn.form_num2(b)}x {self.fn.form_num3(c)}$' # y = ax^2 + bx + c と表記する．

    def range_def(self, x_def_range):
        """ 凡例のための定義域を作成 """        
        x_def_min, x_def_max = x_def_range
        
        return rf'${x_def_min} \leq x \leq {x_def_max}$の範囲' # x_def_min <= x <= x_def_max と表記する

    def range_lin(self, a, b, inequality_type):
        """ 凡例のための1次不等式を作成 """  
        if b == 0:
            return rf'${self.fn.form_num1(a)}x {inequality_type} 0$の範囲'
        else:
            return rf'${self.fn.form_num1(a)}x {self.fn.form_num3(b)} {inequality_type} 0$の範囲'

    def range_quad(self, a, b, c, inequality_type):
        if b == 0 and c == 0:
            return rf'${self.fn.form_num1(a)}x^2 {inequality_type} 0$の範囲'
        elif b == 0:
            return rf'${self.fn.form_num1(a)}x^2 {self.fn.form_num3(c)} {inequality_type} 0$の範囲'
        elif c == 0:
            return rf'${self.fn.form_num1(a)}x^2 {self.fn.form_num2(b)}x {inequality_type} 0$の範囲'
        else:
            return rf'${self.fn.form_num1(a)}x^2 {self.fn.form_num2(b)}x {self.fn.form_num3(c)} {inequality_type} 0$の範囲'


class Point_Labelling:
    """
    点のラベルを印字するクラス
    """
    def __init__(self):
        # 各点関連のサイズ
        self._fontsize = 18 # ラベルのフォントサイズ

    def sol(self, a, p, pt, label):
        """ 解のラベル付け（位置を微調整して表示）"""
        # ラベルの高さ調整（下に凸，上に凸に応じて調整）
        if a > 0:
            va_pos = 'top'
            offset = -1
        else:
            va_pos = 'bottom'
            offset = 1
            
        # 解の座標を表示(表示位置を縦方向に微調整し，さらに横方向をhaにより調整，縦方向をvaにより調整）
        if pt[0] < p:
            plt.text(pt[0], pt[1] + offset*0.25, label, color='red', fontsize=self._fontsize, ha='right', va = va_pos)
        else:
            plt.text(pt[0], pt[1] + offset*0.25, label, color='red', fontsize=self._fontsize, ha='left', va = va_pos)

    def pt(self, a, p, pt, label):
        """ 各点のラベル付け（位置を微調整して表示）"""
        # ラベルの高さ調整（下に凸，上に凸に応じて調整）
        if a > 0:
            va_pos = 'top'
        else:
            va_pos = 'bottom'
        # 各点の座標を表示(表示位置を横方向に微調整し，さらに横方向をhaにより調整，縦方向をvaにより調整）
        if pt[0] < p:
            plt.text(pt[0] - 0.2, pt[1], label, color='red', fontsize=self._fontsize, ha='right', va = va_pos)
        else:
            plt.text(pt[0] + 0.2, pt[1], label, color='red', fontsize=self._fontsize, ha='left', va = va_pos)


class Graph_Decoration:
    """
    グラフの体裁を整えるクラス
    """
    @staticmethod   # 静的メソッドとして定義                  
    def ex(x_range, y_range):
        """ グラフの体裁を設定 """
        #----------(パラメータ追加設定)---------------------
        # x, y軸関連文字のサイズ
        div_labelsize = 14 # 軸目盛のラベルサイズ 
        axis_fontsize = 18 # 軸のフォントサイズ
        
        # 凡例フォントサイズ
        legend_fontsize = 14 
        #---------(パラメータ追加設定おわり)-----------------

        x_min, x_max = x_range
        y_min, y_max = y_range
        
        # 太線の座標軸の設定
        plt.axhline(0, color='black', linewidth=1)  # x 軸を太線で表示
        plt.axvline(0, color='black', linewidth=1)  # y 軸を太線で表示
        
        # 軸ラベルの表示
        plt.text(x_max + 0.5, 0, 'x', va='center', ha='center', fontsize=axis_fontsize) # x軸ラベルの表示
        plt.text(0, y_max + 0.5, 'y', va='center', ha='center', rotation=0, fontsize=axis_fontsize) # y軸ラベルを上側に表示
        
        # グリッドを破線で表示し、整数の目盛りに設定
        plt.grid(True, linestyle='--')
        plt.xticks(np.arange(x_min, x_max + 0.02, 1))
        plt.yticks(np.arange(y_min, y_max + 0.02, 1))
        
        # 凡例をグラフの外に配置
        plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0., fontsize = legend_fontsize)
        
        # x軸を y=0 に、y軸を x=0 に配置
        plt.gca().spines['bottom'].set_position(('data', 0))
        plt.gca().spines['left'].set_position(('data', 0))
        
        # 上端と右端の枠線を非表示にする
        plt.gca().spines['top'].set_color('none')
        plt.gca().spines['right'].set_color('none')
        
        # 目盛を内向きに設定
        plt.tick_params(axis='x', direction='in', length=5, labelsize = div_labelsize)
        plt.tick_params(axis='y', direction='in', length=5, labelsize = div_labelsize)
        
        # 縦横比を等しく設定し、枠の幅を調整可能にする
        plt.gca().set_aspect('equal', adjustable='box')
        
        # グラフの x 軸と y 軸の表示範囲を設定
        plt.xlim(x_min, x_max + 0.02)
        plt.ylim(y_min - 0.02 , y_max + 0.02 )






